#!/usr/bin/env bash
#
# picocompletion-demo Bash Completion
# =======================
#
# Bash completion support for the `picocompletion-demo` command,
# generated by [picocli](https://picocli.info/) version %1$s.
#
# Installation
# ------------
#
# 1. Source all completion scripts in your .bash_profile
#
#   cd $YOUR_APP_HOME/bin
#   for f in $(find . -name "*_completion"); do line=". $(pwd)/$f"; grep "$line" ~/.bash_profile || echo "$line" >> ~/.bash_profile; done
#
# 2. Open a new bash console, and type `picocompletion-demo [TAB][TAB]`
#
# 1a. Alternatively, if you have [bash-completion](https://github.com/scop/bash-completion) installed:
#     Place this file in a `bash-completion.d` folder:
#
#   * /etc/bash-completion.d
#   * /usr/local/etc/bash-completion.d
#   * ~/bash-completion.d
#
# Documentation
# -------------
# The script is called by bash whenever [TAB] or [TAB][TAB] is pressed after
# 'picocompletion-demo (..)'. By reading entered command line parameters,
# it determines possible bash completions and writes them to the COMPREPLY variable.
# Bash then completes the user input if only one entry is listed in the variable or
# shows the options if more than one is listed in COMPREPLY.
#
# References
# ----------
# [1] http://stackoverflow.com/a/12495480/1440785
# [2] http://tiswww.case.edu/php/chet/bash/FAQ
# [3] https://www.gnu.org/software/bash/manual/html_node/The-Shopt-Builtin.html
# [4] http://zsh.sourceforge.net/Doc/Release/Options.html#index-COMPLETE_005fALIASES
# [5] https://stackoverflow.com/questions/17042057/bash-check-element-in-array-for-elements-in-another-array/17042655#17042655
# [6] https://www.gnu.org/software/bash/manual/html_node/Programmable-Completion.html#Programmable-Completion
# [7] https://stackoverflow.com/questions/3249432/can-a-bash-tab-completion-script-be-used-in-zsh/27853970#27853970
#

if [ -n "$BASH_VERSION" ]; then
  # Enable programmable completion facilities when using bash (see [3])
  shopt -s progcomp
elif [ -n "$ZSH_VERSION" ]; then
  # Make alias a distinct command for completion purposes when using zsh (see [4])
  setopt COMPLETE_ALIASES
  alias compopt=complete

  # Enable bash completion in zsh (see [7])
  # Only initialize completions module once to avoid unregistering existing completions.
  if ! type compdef > /dev/null; then
    autoload -U +X compinit && compinit
  fi
  autoload -U +X bashcompinit && bashcompinit
fi

# CompWordsContainsArray takes an array and then checks
# if all elements of this array are in the global COMP_WORDS array.
#
# Returns zero (no error) if all elements of the array are in the COMP_WORDS array,
# otherwise returns 1 (error).
function CompWordsContainsArray() {
  declare -a localArray
  localArray=("$@")
  local findme
  for findme in "${localArray[@]}"; do
    if ElementNotInCompWords "$findme"; then return 1; fi
  done
  return 0
}
function ElementNotInCompWords() {
  local findme="$1"
  local element
  for element in "${COMP_WORDS[@]}"; do
    if [[ "$findme" = "$element" ]]; then return 1; fi
  done
  return 0
}

# The `currentPositionalIndex` function calculates the index of the current positional parameter.
#
# currentPositionalIndex takes three parameters:
# the command name,
# a space-separated string with the names of options that take a parameter, and
# a space-separated string with the names of boolean options (that don't take any params).
# When done, this function echos the current positional index to std_out.
#
# Example usage:
# local currIndex=$(currentPositionalIndex "mysubcommand" "$ARG_OPTS" "$FLAG_OPTS")
function currentPositionalIndex() {
  local commandName="$1"
  local optionsWithArgs="$2"
  local booleanOptions="$3"
  local previousWord
  local result=0

  for i in $(seq $((COMP_CWORD - 1)) -1 0); do
    previousWord=${COMP_WORDS[i]}
    if [ "${previousWord}" = "$commandName" ]; then
      break
    fi
    if [[ "${optionsWithArgs}" =~ ${previousWord} ]]; then
      ((result-=2)) # Arg option and its value not counted as positional param
    elif [[ "${booleanOptions}" =~ ${previousWord} ]]; then
      ((result-=1)) # Flag option itself not counted as positional param
    fi
    ((result++))
  done
  echo "$result"
}

# compReplyArray generates a list of completion suggestions based on an array, ensuring all values are properly escaped.
#
# compReplyArray takes a single parameter: the array of options to be displayed
#
# The output is echoed to std_out, one option per line.
#
# Example usage:
# local options=("foo", "bar", "baz")
# local IFS=$'\n'
# COMPREPLY=($(compReplyArray "${options[@]}"))
function compReplyArray() {
  declare -a options
  options=("$@")
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local i
  local quoted
  local optionList=()

  for (( i=0; i<${#options[@]}; i++ )); do
    # Double escape, since we want escaped values, but compgen -W expands the argument
    printf -v quoted %%q "${options[i]}"
    quoted=\'${quoted//\'/\'\\\'\'}\'

    optionList[i]=$quoted
  done

  # We also have to add another round of escaping to $curr_word.
  curr_word=${curr_word//\\/\\\\}
  curr_word=${curr_word//\'/\\\'}

  # Actually generate completions.
  local IFS=$'\n'
  echo -e "$(compgen -W "${optionList[*]}" -- "$curr_word")"
}

# Bash completion entry point function.
# _complete_picocompletion-demo finds which commands and subcommands have been specified
# on the command line and delegates to the appropriate function
# to generate possible options and subcommands for the last specified subcommand.
function _complete_picocompletion-demo() {
  # Edge case: if command line has no space after subcommand, then don't assume this subcommand is selected (remkop/picocli#1468).
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub1" ];    then _picocli_picocompletion-demo; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub1-alias" ];    then _picocli_picocompletion-demo; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2" ];    then _picocli_picocompletion-demo; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias" ];    then _picocli_picocompletion-demo; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 subsub1" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 sub2child1-alias" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 subsub2" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 sub2child2-alias" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 subsub3" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2 sub2child3-alias" ];    then _picocli_picocompletion-demo_sub2; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias subsub1" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias sub2child1-alias" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias subsub2" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias sub2child2-alias" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias subsub3" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if [ "${COMP_LINE}" = "${COMP_WORDS[0]} sub2-alias sub2child3-alias" ];    then _picocli_picocompletion-demo_sub2alias; return $?; fi

  # Find the longest sequence of subcommands and call the bash function for that subcommand.
  local cmds0=(sub1)
  local cmds1=(sub1-alias)
  local cmds2=(sub2)
  local cmds3=(sub2-alias)
  local cmds4=(sub2 subsub1)
  local cmds5=(sub2 sub2child1-alias)
  local cmds6=(sub2 subsub2)
  local cmds7=(sub2 sub2child2-alias)
  local cmds8=(sub2 subsub3)
  local cmds9=(sub2 sub2child3-alias)
  local cmds10=(sub2-alias subsub1)
  local cmds11=(sub2-alias sub2child1-alias)
  local cmds12=(sub2-alias subsub2)
  local cmds13=(sub2-alias sub2child2-alias)
  local cmds14=(sub2-alias subsub3)
  local cmds15=(sub2-alias sub2child3-alias)

  if CompWordsContainsArray "${cmds15[@]}"; then _picocli_picocompletion-demo_sub2alias_sub2child3alias; return $?; fi
  if CompWordsContainsArray "${cmds14[@]}"; then _picocli_picocompletion-demo_sub2alias_subsub3; return $?; fi
  if CompWordsContainsArray "${cmds13[@]}"; then _picocli_picocompletion-demo_sub2alias_sub2child2alias; return $?; fi
  if CompWordsContainsArray "${cmds12[@]}"; then _picocli_picocompletion-demo_sub2alias_subsub2; return $?; fi
  if CompWordsContainsArray "${cmds11[@]}"; then _picocli_picocompletion-demo_sub2alias_sub2child1alias; return $?; fi
  if CompWordsContainsArray "${cmds10[@]}"; then _picocli_picocompletion-demo_sub2alias_subsub1; return $?; fi
  if CompWordsContainsArray "${cmds9[@]}"; then _picocli_picocompletion-demo_sub2_sub2child3alias; return $?; fi
  if CompWordsContainsArray "${cmds8[@]}"; then _picocli_picocompletion-demo_sub2_subsub3; return $?; fi
  if CompWordsContainsArray "${cmds7[@]}"; then _picocli_picocompletion-demo_sub2_sub2child2alias; return $?; fi
  if CompWordsContainsArray "${cmds6[@]}"; then _picocli_picocompletion-demo_sub2_subsub2; return $?; fi
  if CompWordsContainsArray "${cmds5[@]}"; then _picocli_picocompletion-demo_sub2_sub2child1alias; return $?; fi
  if CompWordsContainsArray "${cmds4[@]}"; then _picocli_picocompletion-demo_sub2_subsub1; return $?; fi
  if CompWordsContainsArray "${cmds3[@]}"; then _picocli_picocompletion-demo_sub2alias; return $?; fi
  if CompWordsContainsArray "${cmds2[@]}"; then _picocli_picocompletion-demo_sub2; return $?; fi
  if CompWordsContainsArray "${cmds1[@]}"; then _picocli_picocompletion-demo_sub1alias; return $?; fi
  if CompWordsContainsArray "${cmds0[@]}"; then _picocli_picocompletion-demo_sub1; return $?; fi

  # No subcommands were specified; generate completions for the top-level command.
  _picocli_picocompletion-demo; return $?;
}

# Generates completions for the options and subcommands of the `picocompletion-demo` command.
function _picocli_picocompletion-demo() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}

  local commands="sub1 sub1-alias sub2 sub2-alias"
  local flag_opts="-V --version -h --help"
  local arg_opts=""

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub1` subcommand.
function _picocli_picocompletion-demo_sub1() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="--num --str --candidates"
  local str2_option_args=("aaa" "bbb" "ccc") # --candidates values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    --num)
      return
      ;;
    --str)
      return
      ;;
    --candidates)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${str2_option_args[@]}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub1-alias` subcommand.
function _picocli_picocompletion-demo_sub1alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="--num --str --candidates"
  local str2_option_args=("aaa" "bbb" "ccc") # --candidates values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    --num)
      return
      ;;
    --str)
      return
      ;;
    --candidates)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${str2_option_args[@]}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2` subcommand.
function _picocli_picocompletion-demo_sub2() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands="subsub1 sub2child1-alias subsub2 sub2child2-alias subsub3 sub2child3-alias"
  local flag_opts=""
  local arg_opts="--num2 --directory -d"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    --num2)
      return
      ;;
    --directory|-d)
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -f -- "${curr_word}" ) ) # files
      return $?
      ;;
  esac
  local possibilities_pos_param_args=("Aaa" "Bbb" "Ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${possibilities_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2-alias` subcommand.
function _picocli_picocompletion-demo_sub2alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands="subsub1 sub2child1-alias subsub2 sub2child2-alias subsub3 sub2child3-alias"
  local flag_opts=""
  local arg_opts="--num2 --directory -d"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    --num2)
      return
      ;;
    --directory|-d)
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -f -- "${curr_word}" ) ) # files
      return $?
      ;;
  esac
  local possibilities_pos_param_args=("Aaa" "Bbb" "Ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2-alias" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${possibilities_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub1` subcommand.
function _picocli_picocompletion-demo_sub2_subsub1() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-h --host"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -h|--host)
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -A hostname -- "${curr_word}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child1-alias` subcommand.
function _picocli_picocompletion-demo_sub2_sub2child1alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-h --host"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -h|--host)
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -A hostname -- "${curr_word}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub2` subcommand.
function _picocli_picocompletion-demo_sub2_subsub2() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-u --timeUnit -t --timeout"
  local timeUnit_option_args=("%2$s") # --timeUnit values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -u|--timeUnit)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${timeUnit_option_args[@]}" ) )
      return $?
      ;;
    -t|--timeout)
      return
      ;;
  esac
  local str2_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "subsub2" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${str2_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child2-alias` subcommand.
function _picocli_picocompletion-demo_sub2_sub2child2alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-u --timeUnit -t --timeout"
  local timeUnit_option_args=("%2$s") # --timeUnit values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -u|--timeUnit)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${timeUnit_option_args[@]}" ) )
      return $?
      ;;
    -t|--timeout)
      return
      ;;
  esac
  local str2_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2child2-alias" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${str2_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub3` subcommand.
function _picocli_picocompletion-demo_sub2_subsub3() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}

  local commands=""
  local flag_opts=""
  local arg_opts=""
  local cands_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "subsub3" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${cands_pos_param_args[@]}" )
    elif (( currIndex >= 1 && currIndex <= 2 )); then
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -f -- "${curr_word}" ) # files
    elif (( currIndex >= 3 && currIndex <= 2147483647 )); then
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -A hostname -- "${curr_word}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child3-alias` subcommand.
function _picocli_picocompletion-demo_sub2_sub2child3alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}

  local commands=""
  local flag_opts=""
  local arg_opts=""
  local cands_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2child3-alias" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${cands_pos_param_args[@]}" )
    elif (( currIndex >= 1 && currIndex <= 2 )); then
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -f -- "${curr_word}" ) # files
    elif (( currIndex >= 3 && currIndex <= 2147483647 )); then
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -A hostname -- "${curr_word}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub1` subcommand.
function _picocli_picocompletion-demo_sub2alias_subsub1() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-h --host"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -h|--host)
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -A hostname -- "${curr_word}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child1-alias` subcommand.
function _picocli_picocompletion-demo_sub2alias_sub2child1alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-h --host"

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -h|--host)
      type compopt &>/dev/null && compopt -o filenames
      COMPREPLY=( $( compgen -A hostname -- "${curr_word}" ) )
      return $?
      ;;
  esac

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub2` subcommand.
function _picocli_picocompletion-demo_sub2alias_subsub2() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-u --timeUnit -t --timeout"
  local timeUnit_option_args=("%2$s") # --timeUnit values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -u|--timeUnit)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${timeUnit_option_args[@]}" ) )
      return $?
      ;;
    -t|--timeout)
      return
      ;;
  esac
  local str2_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "subsub2" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${str2_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child2-alias` subcommand.
function _picocli_picocompletion-demo_sub2alias_sub2child2alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}
  local prev_word=${COMP_WORDS[COMP_CWORD-1]}

  local commands=""
  local flag_opts=""
  local arg_opts="-u --timeUnit -t --timeout"
  local timeUnit_option_args=("%2$s") # --timeUnit values

  type compopt &>/dev/null && compopt +o default

  case ${prev_word} in
    -u|--timeUnit)
      local IFS=$'\n'
      COMPREPLY=( $( compReplyArray "${timeUnit_option_args[@]}" ) )
      return $?
      ;;
    -t|--timeout)
      return
      ;;
  esac
  local str2_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2child2-alias" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${str2_pos_param_args[@]}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `subsub3` subcommand.
function _picocli_picocompletion-demo_sub2alias_subsub3() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}

  local commands=""
  local flag_opts=""
  local arg_opts=""
  local cands_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "subsub3" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${cands_pos_param_args[@]}" )
    elif (( currIndex >= 1 && currIndex <= 2 )); then
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -f -- "${curr_word}" ) # files
    elif (( currIndex >= 3 && currIndex <= 2147483647 )); then
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -A hostname -- "${curr_word}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Generates completions for the options and subcommands of the `sub2child3-alias` subcommand.
function _picocli_picocompletion-demo_sub2alias_sub2child3alias() {
  # Get completion data
  local curr_word=${COMP_WORDS[COMP_CWORD]}

  local commands=""
  local flag_opts=""
  local arg_opts=""
  local cands_pos_param_args=("aaa" "bbb" "ccc") # 0-0 values

  if [[ "${curr_word}" == -* ]]; then
    COMPREPLY=( $(compgen -W "${flag_opts} ${arg_opts}" -- "${curr_word}") )
  else
    local positionals=""
    local currIndex
    currIndex=$(currentPositionalIndex "sub2child3-alias" "${arg_opts}" "${flag_opts}")
    if (( currIndex >= 0 && currIndex <= 0 )); then
      positionals=$( compReplyArray "${cands_pos_param_args[@]}" )
    elif (( currIndex >= 1 && currIndex <= 2 )); then
      local IFS=$'\n'
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -f -- "${curr_word}" ) # files
    elif (( currIndex >= 3 && currIndex <= 2147483647 )); then
      type compopt &>/dev/null && compopt -o filenames
      positionals=$( compgen -A hostname -- "${curr_word}" )
    fi
    local IFS=$'\n'
    COMPREPLY=( $(compgen -W "${commands// /$'\n'}${IFS}${positionals}" -- "${curr_word}") )
  fi
}

# Define a completion specification (a compspec) for the
# `picocompletion-demo`, `picocompletion-demo.sh`, and `picocompletion-demo.bash` commands.
# Uses the bash `complete` builtin (see [6]) to specify that shell function
# `_complete_picocompletion-demo` is responsible for generating possible completions for the
# current word on the command line.
# The `-o default` option means that if the function generated no matches, the
# default Bash completions and the Readline default filename completions are performed.
complete -F _complete_picocompletion-demo -o default picocompletion-demo picocompletion-demo.sh picocompletion-demo.bash
